package com.maroontress.sanctuary;

import java.io.File;
import java.io.PrintWriter;
import org.w3c.dom.Element;

/**
   饹Ǥ
*/
public class Class extends Group {

    /** 󥯥롼ɤΥå */
    private Includes includes;

    /** ᥽åɤΥå */
    private RegularMethods methods;

    /** ֥ȥ饯ȥ᥽åɤΥå */
    private AbstractMethods abstractMethods;

    /** ޥΥå */
    private Macros macros;

    /** Υå */
    private Typedefs typedefs;

    /** ХѿΥå */
    private GlobalVariables globalVariables;

    /** Υ饹˽°륳ѥ */
    private Compounds compounds;

    /** ݡȤ륿Υꥹ */
    private ImportedTags importedTags;

    /** ݡȤXClassΥꥹ */
    private ImportedXClass importedXClass;

    /** ̾Ÿλ */
    private boolean hasNS;

    /**
       饹ޤ

       @param elem class
       @param filename XMLΥե̾
    */
    public Class(final Element elem, final String filename) {
	super(elem, filename);

	IDManager idManager = IDManager.getInstance();
	String packageName = idManager.getIDAttribute(elem, "package");
	String name = idManager.getIDAttribute(elem, "name");
	setLocation(new Location(packageName, name));

	ChildrenFilter c = getChildrenFilter();
	hasNS = (c.getDirectChildrenByTagName("namespace").getLength() > 0);

	compounds = new Compounds(this);
	importedTags = new ImportedTags();
	importedXClass = new ImportedXClass();

        includes = new Includes(c.getDirectChildrenByTagName("include"));

	compounds.addExports(c.getDirectChildrenByTagName("export"));
	compounds.addToplevelCompunds(
	    c.getDirectChildrenByTagName("struct|union|enum"));

	methods = new RegularMethods(
	    this, c.getDirectChildrenByTagName("method"));
	abstractMethods = new AbstractMethods(
	    this, c.getDirectChildrenByTagName("abstractmethod"));
	macros = new Macros(this, c.getDirectChildrenByTagName("define"));
	typedefs = new Typedefs(this, c.getDirectChildrenByTagName("typedef"));
	globalVariables = new GlobalVariables(
	    this, c.getDirectChildrenByTagName("globalvariable"));

	/* importǤɲäȾĹimportǤθ */
	importedTags.addNodeList(this, c.getDirectChildrenByTagName("import"));
    }

    /**
       {@inheritDoc}
    */
    @Override protected boolean isNamespaceExpanded() {
	return hasNS;
    }

    /**
       {@inheritDoc}
    */
    @Override public void printCode() {
	String packageName = getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";
	String importMacro = packageName + "_" + name + "_IMPORT";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	includes.printCode(wr);
	importedXClass.printCode(wr);
	macros.printCode(wr);
	importedTags.printCode(wr);
	typedefs.printCode(wr);
	compounds.printCode(wr);
	methods.printCode(wr);
	globalVariables.printCode(wr);

	if (isNamespaceExpanded()) {
	    wr.println("");
	    wr.println("#ifdef " + importMacro);
	    printImportMacro(wr);
	    wr.println("#endif /* " + importMacro + " */");
	}

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();
    }

    /**
       {@inheritDoc}
    */
    @Override protected String getType() {
	return "class";
    }

    /**
       {@inheritDoc}
    */
    @Override protected String getTexType() {
	return "class";
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtmlBodyHeader(final PrintWriter wr) {
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtmlUsage(final PrintWriter wr) {
	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";
	if (isNamespaceExpanded()) {
	    wr.println("#define " + s + "_IMPORT<br>");
	}
	wr.println("#include &lt;" + h + "&gt;<br>");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtml(final PrintWriter wr) {
	macros.printSummary(wr);
	typedefs.printSummary(wr);
	compounds.printHtmlSummary(wr);
	methods.printSummary(wr);
	abstractMethods.printSummary(wr);
	globalVariables.printSummary(wr);

	macros.printDetail(wr);
	typedefs.printDetail(wr);
	compounds.printHtmlDetail(wr);
	methods.printDetail(wr);
	abstractMethods.printDetail(wr);
	globalVariables.printDetail(wr);
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTexUsage(final PrintWriter wr) {
	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";
	wr.println("\\nopagebreak");
	wr.println("\\begin{Ssyntax}%");
	if (isNamespaceExpanded()) {
	    wr.println(Tex.safe("#define " + s + "_IMPORT") + "\\\\");
	}
	wr.println(Tex.safe("#include <" + h + ">") + "\\\\");
	wr.println("\\end{Ssyntax}");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTex(final PrintWriter wr) {
	macros.printTexDetail(wr);
	typedefs.printTexDetail(wr);
	compounds.printTexDetail(wr);
	methods.printTexDetail(wr);
	abstractMethods.printTexDetail(wr);
	globalVariables.printTexDetail(wr);
    }

    /**
       {@inheritDoc}
    */
    @Override public void printStub() {
	printStubHeader();
	printStubCode();
    }

    /**
       ֤Υإåեޤ
    */
    private void printStubHeader() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("");
	wr.println("#include <" + s + ".h>");

	methods.printStubHeader(wr);
	globalVariables.printStubHeader(wr);

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();
    }

    /**
       ֤Υɤޤ
    */
    private void printStubCode() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".c");

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("#include <stdlib.h>");
	wr.println("#include <stdarg.h>");
	wr.println("#include <" + s + ".h>");

	methods.printStubCode(wr);
	globalVariables.printStubCode(wr);
	wr.close();
    }

    /**
       {@inheritDoc}
    */
    @Override public void importTags(final String s) {
	importedTags.add(s);
    }

    /**
       {@inheritDoc}
    */
    @Override public void importTag(final String type, final String tag) {
	importedTags.add(type, tag);
    }

    /**
       {@inheritDoc}
    */
    @Override protected void importXClass(final String fqcn) {
	importedXClass.add(fqcn);
    }

    /**
       {@inheritDoc}

       s "@" ξ prefixʳξprefixs "_" Ϣ
       ʸ
    */
    @Override protected String prependPrefix(final String prefix,
					     final String s) {
	return (s.equals("@")) ? prefix : prefix + "_" + s;
    }

    /**
       {@inheritDoc}
    */
    @Override public void addCompound(final Compound compound,
				      final String type) {
	compounds.addCompound(compound, type);
    }
}
