package com.maroontress.sanctuary.xclass;

import com.maroontress.sanctuary.ChildrenFilter;
import com.maroontress.sanctuary.Compound;
import com.maroontress.sanctuary.Compounds;
import com.maroontress.sanctuary.GlobalVariables;
import com.maroontress.sanctuary.Group;
import com.maroontress.sanctuary.IDManager;
import com.maroontress.sanctuary.ImportedTags;
import com.maroontress.sanctuary.ImportedXClass;
import com.maroontress.sanctuary.Includes;
import com.maroontress.sanctuary.Macros;
import com.maroontress.sanctuary.NodeException;
import com.maroontress.sanctuary.Options;
import com.maroontress.sanctuary.QualifiedName;
import com.maroontress.sanctuary.Typedefs;
import com.maroontress.sanctuary.Tex;
import java.io.File;
import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.List;
import org.w3c.dom.Element;

/**
   ĥǽʥ饹Ǥ
*/
public class XClass extends Group {

    /** ֤̾ΰ */
    private XLocation xlocation;

    /** 󥯥롼ɤΥå */
    private Includes includes;

    /** եɤΥå */
    private XFields fields;

    /** ᥽åɤΥå */
    private XRegularMethods methods;

    /** ޥΥå */
    private Macros macros;

    /** Υå */
    private Typedefs typedefs;

    /** ХѿΥå */
    private GlobalVariables globalVariables;

    /** Υ饹˽°륳ѥ */
    private Compounds compounds;

    /** ݡȤ륿Υꥹ */
    private ImportedTags importedTags;

    /** ݡȤXClassΥꥹ */
    private ImportedXClass importedXClass;

    /** ̾Ÿλ */
    private boolean hasNS;

    /** ѾطΥꥹ */
    private List<XLocation> inheritanceList;

    /**
       ĥǽʥ饹ޤ

       @param elem class
       @param filename XMLΥե̾
    */
    public XClass(final Element elem, final String filename) {
	super(elem, filename);

	IDManager idManager = IDManager.getInstance();
	xlocation = XLocation.intern(idManager.getFQCNAttribute(elem, "name"));
	xlocation.setKnown();
        String parent = idManager.getFQCNImpliedAttribute(elem, "extends");
        if (parent == null) {
            parent = Options.getInstance().getBaseXClass();
	    if (parent != null && parent.equals(xlocation.getFQCN())) {
		parent = null;
	    }
        }
        if (parent != null) {
	    XLocation loc = XLocation.intern(parent);
	    xlocation.setParent(loc);
	    while (loc != null && loc != xlocation) {
		loc = loc.getParent();
	    }
	    if (loc != null) {
		throw new NodeException(elem.getAttributeNode("extends"),
					"inheritance loop found.");
	    }
	}
	setLocation(xlocation);

	QualifiedName name = QualifiedName.intern(xlocation.getPrefix(),
						  xlocation.getSimplePrefix());
	addKeyword(QualifiedName.prepend("union ", name));

	ChildrenFilter c = getChildrenFilter();
	compounds = new Compounds(this);
	importedTags = new ImportedTags();
	importedXClass = new ImportedXClass();

        includes = new Includes(c.getDirectChildrenByTagName("include"));

	compounds.addExports(c.getDirectChildrenByTagName("export"));
	compounds.addToplevelCompunds(
	    c.getDirectChildrenByTagName("struct|union|enum"));

        fields = new XFields(
	    this, c.getDirectChildrenByTagName("field"));
	methods = new XRegularMethods(
	    this, c.getDirectChildrenByTagName("instancemethod|staticmethod"));
	macros = new Macros(this, c.getDirectChildrenByTagName("define"));
	typedefs = new Typedefs(this, c.getDirectChildrenByTagName("typedef"));
	globalVariables = new GlobalVariables(
	    this, c.getDirectChildrenByTagName("globalvariable"));

	/* importǤɲäȾĹimportǤθ */
	importedTags.addNodeList(this, c.getDirectChildrenByTagName("import"));
	inheritanceList = null;
    }

    /**
       {@inheritDoc}
    */
    @Override protected boolean isNamespaceExpanded() {
	return true;
    }

    /**
       ѾطɽΥꥹȤޤ

       ΥꥹȤǤϡƬ쥯饹ΤΡǸ
       ǤʬȤΥ饹ΤΤȤ¤Ӥޤ

       @return Υꥹ
    */
    private List<XLocation> getInheritanceList() {
	if (inheritanceList != null) {
	    return inheritanceList;
	}
	LinkedList <XLocation> list = new LinkedList<XLocation>();

	list.add(xlocation);
	for (XLocation loc = xlocation.getParent();
             loc != null;
             loc = loc.getParent()) {
	    list.add(0, loc);
        }
	inheritanceList = list;
	return inheritanceList;
    }

    /**
       Υإåեޤ
    */
    private void printImplCode() {
	final XLocation parent = xlocation.getParent();
	String packageName = getPackageName() + "_impl";
	String dirName = packageName.replace('_', File.separatorChar);
	String name = getName();
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
        String prefix = xlocation.getImplPrefix();
        String includeMacro = prefix + "_H";
        String importMacro = xlocation.getPrefix() + "_IMPORT";

        wr.println("#ifndef " + includeMacro);
        wr.println("#define " + includeMacro);

        includes.printCode(wr);

        wr.println("");
        for (XLocation loc = parent; loc != null; loc = loc.getParent()) {
            wr.println("#include <" + loc.getImplPath() + ".h>");
        }
        wr.println("#include <" + xlocation.getPath() + ".h>");

        wr.println("");
        wr.println("struct " + xlocation.getImplPrefix() + " {");
	if (parent != null) {
	    wr.println("    union {");
	    for (XLocation loc = parent; loc != null; loc = loc.getParent()) {
		wr.println("        struct " + loc.getImplPrefix() + " "
			   + loc.getClassMember() + ";");
	    }
	    wr.println("    } super;");
	}
        fields.printCode(wr);
        wr.println("};");

        wr.println("");
        wr.println("#ifdef " + importMacro);
        wr.println("#define impl_" + xlocation.getClassName() + " "
                   + xlocation.getImplPrefix());
        wr.println("#endif /* " + importMacro + " */");

        wr.println("");
        wr.println("#endif /* !" + includeMacro + " */");
        wr.close();
    }

    /**
       {@inheritDoc}
    */
    @Override public void printCode() {
	final XLocation parent = xlocation.getParent();
	String packageName = getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";
	String importMacro = packageName + "_" + name + "_IMPORT";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	includes.printCode(wr);
	importedXClass.printCode(wr);
	macros.printCode(wr);
	importedTags.printCode(wr);
	typedefs.printCode(wr);
	compounds.printCode(wr);

        wr.println("");
        for (XLocation loc = xlocation; loc != null; loc = loc.getParent()) {
            wr.println("struct " + loc.getImplPrefix() + ";");
        }

	wr.println("");
        wr.println("union " + xlocation.getPrefix() + " {");
	if (parent != null) {
	    wr.println("    union {");
	    for (XLocation loc = parent; loc != null; loc = loc.getParent()) {
		wr.println("        struct " + loc.getImplPrefix() + " *"
			   + loc.getClassMember() + ";");
	    }
	    wr.println("    } super;");
	}
        wr.println("    struct " + xlocation.getImplPrefix() + " *impl;");
        wr.println("} __attribute__((__transparent_union__));");

	methods.printCode(wr);
	globalVariables.printCode(wr);

	wr.println("");
	wr.println("#ifdef " + importMacro);
	wr.println("#define " + xlocation.getSimplePrefix()
		   + " " + xlocation.getPrefix());
	printImportMacro(wr);
	wr.println("#endif /* " + importMacro + " */");

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();

        printImplCode();
    }

    /**
       {@inheritDoc}
    */
    @Override protected String getType() {
	return "extensible class";
    }

    /**
       {@inheritDoc}
    */
    @Override protected String getTexType() {
	return "xclass";
    }

    /**
       饹ηѾطλHTMLǽϤޤ

       @param wr 饤
       @param d Υ٥
    */
    private void printHtmlInheritanceIndent(final PrintWriter wr, final int d) {
	for (int k = 1; k < d; ++k) {
	    wr.print("&nbsp;&nbsp;&nbsp;&nbsp;");
	}
	wr.print("&#x2514;&nbsp;");
    }

    /**
       Υ󥫡ʸHTMLǼޤ

       @param loc 
       @return HTMLΥ󥫡ʸ
    */
    private String getHtmlInheritanceAnchor(final XLocation loc) {
	if (loc == xlocation) {
	    return "<b>" + loc.getPrefix() + "</b>";
	}
	if (!loc.isKnown()) {
	    return loc.getPrefix();
	}
	return "<a href=\"" + loc.getPrefix() + ".html\">"
	    + loc.getPrefix() + "</a>";
    }

    /**
       饹ηѾطHTMLǽϤޤ

       @param wr 饤
    */
    private void printHtmlInheritance(final PrintWriter wr) {
	XLocation[] loc = getInheritanceList().toArray(new XLocation[0]);

	wr.println(getHtmlInheritanceAnchor(loc[0]) + "<br>");
	for (int k = 1; k < loc.length; ++k) {
	    printHtmlInheritanceIndent(wr, k);
	    wr.println(getHtmlInheritanceAnchor(loc[k]) + "<br>");
	}
	wr.println("<br>");

	loc = xlocation.getChildren();
	if (loc.length == 0) {
	    return;
	}
	wr.println("<dl><b>Direct Known Subclasses:</b><dd>");
	String comma = "";
	for (XLocation x : loc) {
	    wr.println(comma + getHtmlInheritanceAnchor(x));
	    comma = ", ";
	}
	wr.println("</dd></dl>");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtmlBodyHeader(final PrintWriter wr) {
	wr.println("<a name=\"union " + xlocation.getPrefix() + "\"/>");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtmlUsage(final PrintWriter wr) {
	printHtmlInheritance(wr);

	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";

	wr.println("#define " + s + "_IMPORT<br>");
	wr.println("#include &lt;" + h + "&gt;<br>");

	s = getPackageName() + "_impl_" + getName();
	h = s.replace('_', File.separatorChar) + ".h";
	wr.println("#include &lt;" + h + "&gt;<br>");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtml(final PrintWriter wr) {
	macros.printSummary(wr);
	typedefs.printSummary(wr);
        fields.printSummary(wr);
	compounds.printHtmlSummary(wr);
	methods.printSummary(wr);
	globalVariables.printSummary(wr);

	macros.printDetail(wr);
	typedefs.printDetail(wr);
        fields.printDetail(wr);
	compounds.printHtmlDetail(wr);
	methods.printDetail(wr);
	globalVariables.printDetail(wr);
    }

    /**
       饹ηѾطλTeXǽϤޤ

       @param wr 饤
       @param d Υ٥
    */
    private void printTexInheritanceIndent(final PrintWriter wr, final int d) {
	if (d > 1) {
	    wr.print("\\hspace{" + (d - 1) + "zw}");
	}
	wr.print("\u2514");
    }

    /**
       Υ󥫡ʸHTMLǼޤ

       @param loc 
       @return HTMLΥ󥫡ʸ
    */
    private String getTexInheritanceAnchor(final XLocation loc) {
	String s = Tex.safe(loc.getPrefix());

	if (loc == xlocation) {
	    return "{\\bfseries " + s + "}";
	}
	if (!loc.isKnown()) {
	    return s;
	}
	String ns = loc.getPackageName() + "." + loc.getClassName();
	return "\\Sanchor{:" +  Tex.safeTarget(ns) + ":}{" + s + "}";
    }

    /**
       饹ηѾطTeXǽϤޤ

       @param wr 饤
    */
    private void printTexInheritance(final PrintWriter wr) {
	XLocation[] loc = getInheritanceList().toArray(new XLocation[0]);

	wr.println(getTexInheritanceAnchor(loc[0]) + "\\\\");
	for (int k = 1; k < loc.length; ++k) {
	    printTexInheritanceIndent(wr, k);
	    wr.println(getTexInheritanceAnchor(loc[k]) + "\\\\");
	}
	wr.println("");

	loc = xlocation.getChildren();
	if (loc.length == 0) {
	    return;
	}
	wr.println("\\SdirectKnownSubclasses{");
	String comma = "";
	for (XLocation x : loc) {
	    wr.println(comma + getTexInheritanceAnchor(x));
	    comma = ", ";
	}
	wr.println("}");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTexUsage(final PrintWriter wr) {
	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";
	String ns = getPackageName() + "." + getName();
	String texLabel = Tex.safeTarget(ns + ".union " + s);

	wr.println("\\Slabel{:" + texLabel  + ":}");
	wr.println("\\nopagebreak");

	wr.println("\\begin{Ssyntax}%");
	printTexInheritance(wr);
	wr.println("\\end{Ssyntax}");

	wr.println("\\begin{Ssyntax}%");
	wr.println(Tex.safe("#define " + s + "_IMPORT") + "\\\\");
	wr.println(Tex.safe("#include <" + h + ">") + "\\\\");

	s = getPackageName() + "_impl_" + getName();
	h = s.replace('_', File.separatorChar) + ".h";
	wr.println(Tex.safe("#include <" + h + ">") + "\\\\");
	wr.println("\\end{Ssyntax}");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTex(final PrintWriter wr) {
	macros.printTexDetail(wr);
	typedefs.printTexDetail(wr);
        fields.printTexDetail(wr);
	compounds.printTexDetail(wr);
	methods.printTexDetail(wr);
	globalVariables.printTexDetail(wr);
    }

    /**
       {@inheritDoc}
    */
    @Override public void printStub() {
	printStubHeader();
	printStubCode();
    }

    /**
       ֤Υإåեޤ
    */
    private void printStubHeader() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("");
	wr.println("#include <" + s + ".h>");

	methods.printStubHeader(wr);
	globalVariables.printStubHeader(wr);

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();
    }

    /**
       ֤Υɤޤ
    */
    private void printStubCode() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".c");

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("#include <stdlib.h>");
	wr.println("#include <stdarg.h>");
	wr.println("#include <" + s + ".h>");

	methods.printStubCode(wr);
	globalVariables.printStubCode(wr);
	wr.close();
    }

    /**
       {@inheritDoc}
    */
    @Override public void importTags(final String s) {
	importedTags.add(s);
    }

    /**
       {@inheritDoc}
    */
    @Override public void importTag(final String type, final String tag) {
	importedTags.add(type, tag);
    }

    /**
       {@inheritDoc}
    */
    @Override protected void importXClass(final String fqcn) {
	if (xlocation.getFQCN().equals(fqcn)) {
	    return;
	}
	importedXClass.add(fqcn);
    }

    /**
       {@inheritDoc}

       prefixs "_" Ϣ뤷ʸ
    */
    @Override protected String prependPrefix(final String prefix,
					     final String s) {
	return prefix + "_" + s;
    }

    /**
       {@inheritDoc}
    */
    @Override public void addCompound(final Compound compound,
					 final String type) {
	compounds.addCompound(compound, type);
    }
}
